/*---------------------------------------------------------------------------------------
 * T.Fountaine
 * October 12, 2012
 *
 * This application demonstrates a method to incorporate a scripting language with
 * the Netburner MOD54415 microcontroller module.
 *
 * The scripting language used is Angelscript which is a subset of the C/C++ language.
 * www.angelcode.com/angelscript
 *
 * The application includes a javascript/jquery UI web based IDE to edit and download scripts.
 * The scripts can be saved and loaded from an SD flash memory card.
 * The Javascript based editor is CodeMirror from www.codemirror.net
 *
 * The jquery UI layout manager that allow resizing the web panes was found at
 * http://layout.jquery-dev.net/
 *
 * The file explore is an AJAX based jquery extension from
 * http://www.abeautifulsite.net/blog/2008/03/jquery-file-tree/
 *
 *
 * A small subset of the Netburner API was registered in the script engine to
 * demonstrate the basic method of extending the scripting engine.
 * The functions include (print, setPin, resetPin, OSTimeDly, hrOSTimeDly)
 *
 * License: This application is not confined by any license.  The author does no provide
 * any guarantee to the correctness of the code.  Please follow the license
 * restrictions of all the added libraries.  Most (if not all) of them are MIT based licenses
 *
 * --------------------------------------------------------------------------------------
 * --------- Note: You must have a valid Netburner development license  -----------------
 * --------------------------------------------------------------------------------------
 *
 ---------------------------------------------------------------------------------------*/


#include "predef.h"
#include <stdio.h>
#include <ctype.h>
#include <startnet.h>
#include <autoupdate.h>
#include <multipartpost.h>
#include <dhcpclient.h>
#include <taskmon.h>
#include "cardtype.h"
#include <effs_fat/fat.h>
#ifdef USE_MMC
#include <effs_fat/mmc_mcf.h>
#endif
#ifdef USE_CFC
#include <effs_fat/cfc_mcf.h>
#endif
#include "FileSystemUtils.h"


#include <angelscript.h>
#include <HiResTimer.h>
#include <string>

#include <ucos.h>
#include <pins.h>
#include <Script.h>

extern "C"
{
   void UserMain( void *pd );
}

const char *AppName = "AngelScript with JQuery based IDE";
double TotalTime = 0;

//Script is a cpp wrapper around the Angelscript library
Script script;

//===== Implement critical section with OSCritEnter() & OSCritLeave() =====
//===== The critical section is lock during script upload from the GUI=====
//===== to prevent calling during compilation                         =====
OS_CRIT cs; // The critical section identifer


////////////////////////////////////////////////////////////////
//Global Variables that will be registered in the script engine
int value = 1000;
////////////////////////////////////////////////////////////////

// Pre Declare our function for HTML form post code.
void RegisterPost();

// Entry point of the application
void UserMain( void *pd )
{
	HiResTimer *timer;
   InitializeStack();

   if ( EthernetIP == 0 )
   {
      iprintf( "Trying DHCP\r\n" );
      GetDHCPAddress();
      iprintf( "DHCP assigned the IP address of :" );
      ShowIP( EthernetIP );
      iprintf( "\r\n" );
   }
   else
   {
      iprintf( "Static IP address set to :" );
      ShowIP( EthernetIP );
      iprintf( "\r\n" );
   }

   const BYTE PinNumber[8] = { 15, 16, 18, 23, 17, 19, 20, 24 };

   for ( int i = 0; i < 8; i++ )
   {
	   J2[PinNumber[i]].function( PIN_GPIO );
	   if(i%2 ==0) J2[PinNumber[i]] =0;
   }


   timer = HiResTimer::getHiResTimer(0);

   //Initialize the critical section used to protect the script engine
   OSCritInit( &cs );

   //Registration of the global variable "value" in the script engine
   script.RegisterGlobalVariable("int value",(void*) &value);

   //OS startup functions
   OSChangePrio( MAIN_PRIO );
   EnableAutoUpdate();
   StartHTTP();
   EnableTaskMonitor();

   //Setup the SD Flash memory card
   f_enterFS();
   InitExtFlash();  // Initialize the CFC or SD/MMC external flash drive
   DisplayEffsSpaceStats();  // Display file space usage

   // Since main will no longer be accessing the file system we can release it
   f_releaseFS();


   /*
   // Enable multi part forms posting with a 500K max data block size
   if ( !EnableMultiPartForms( 500000 ) )
   {
      iprintf( "EnableMultiPartForms() initialization failed\r\n" );
   }
   */

   //Register an HTML post handler
   RegisterPost();

   //Never exits
   while ( 1 )
   {
     // OSTimeDly( TICKS_PER_SECOND );
      OSTimeDly( 1 );

        //Lock the critical section and run the script
	   	OSCritEnter( &cs, 0 );
	   	// Initialize the timer
	   		//timer->init();
	   	    //timer->start();
	   	    script.Run("One_Hz");
	   	    //TotalTime = timer->readTime();
	   	    //timer->stopClear();
		OSCritLeave( &cs );

   }

   //This will never execute
   //Unmount the SD flash card
   UnmountExtFlash();


}
